(**********************************************************************
*                        DiskIteration                              *
***********************************************************************
*  DiskIteration obtains the solution to problem of minimizing over  
*  h in Hinfinity the supnorm of g + 2 re[Sum[a[j]h[j]]] + Sum[H[j] Abs[h[j]]^2    
*  It works by completing the square to obtain a 'perfect square' quadratic
*  which is of the form                                               
*  w ( |k|^2 + 2 Re(k h) + |h|^2 ) = w | Conjugate[k] + h |^2.        
*  Then, it calls Nehari[W,K] repeatedly to get a solution.

   Examples:
	n = 32;
	one = Table[1.,{n}];
	e = Table[N[Exp[2 Pi I (i-1)/n]],{i,1,n}];

	gk1 = Abs[1/(e-0.5)]^2;

	gk2 =  Abs[e + 0.25] + Abs[1/(e-0.5)]^2;

	ak1 = {e/(1.-0.5 e)};

	hk1 = {one};

	ak2 = {e/(1.-0.5 e), (e+2)/(e^2 + e + 4)}

	hk2 = {one, Abs[e^3 -  e +  3]};

	testa := 
	Timing[DiskIteration[gk1,ak1,hk1,DiskIterationDiagnostics->2];]

	testb := 
	Timing[DiskIteration[gk2,ak1,hk1,DiskIterationDiagnostics->2];]

	testc := 
	Timing[DiskIteration[gk2,ak2,hk2,DiskIterationDiagnostics->2];]
***********************************************************************)


BeginPackage["DiskIteration`",{"DualExtremal`","AnoptUtilities`"}];

DiskIteration::usage=" \n
   DiskIteration[g,a,H] Solves the sup norm optimization problem \n
   inf_h sup_e Sum_j  g + 2 Re[ a[[j]] h[[j]] ] + H[[j]] Abs[h[[j]]]^2 \n
   where h is in Hinfinity, e is in the circle, 1 <= j <= ndim, \n
   g is a discrete real valued function (i.e. a list), \n
   a is an ndim-long list of discrete complex valued functions, \n
   H is an ndim-long list of discrete positive valued functions.";
DiskIterationMaxIts::usage=" \n
   Option for DiskIteration.  An integer, it specifies a maximum number \n
   of iterations of the algotithm. ";
DiskIterationDiagnostics::usage=" \n
   Option for DiskIteration.  An integer, it specifies the amount of  \n
   information on the run to be written. ";
DiskIterationTol::usage=" \n
   Option for DiskIteration.  A positive real, specifies the tolerance for \n
   stopping the iteration. ";
DiskIterationOutputChannel::usage=" \n
   Option for DiskIteration.   Specifies the output channel(s) for writing \
   Controled by diag.";
DiskIterationDebugMe::usage=" \n"
Debugger::location::usage = " \n
    Message for DiskIterationDebugMe option.  Function COMMENT[] controls it "
CheckCode::zerodivide::usage = " \n
    Message to check for division by zero"
CheckCode::withintolerance::usage = " \n
    Message to check if within a tolerance"

Options[DiskIteration] = 
		{ DualExtremal`DualMaxIts ->15,
		  DualExtremal`DualTol -> 0.0000001,
		  DualExtremal`DualCircleGrid ->Automatic,
		  DualExtremal`DualSingularFunction->Automatic,
		  AnoptUtilities`HarrisSFMaxIts->10,
		  AnoptUtilities`HarrisSFToler->1.0,
		  DiskIterationMaxIts->15,
		  DiskIterationDiagnostics->0,
		  DiskIterationTol -> 0.0001,
		  DiskIterationOutputChannel->OutputStream["stdout",1],
                  DiskIterationDebugMe->False};


Begin["DiskIteration`Private`"]

DiskIteration[Gk_,Ak_,Hk_,opts___Rule]:=
   Module[{update,errors,

        alpha,overflow,lowdelt,highdelt,deiterations,
	tauratio,W,K,squarecomplete,lowTau,highTau,oldTau,
	highTau2,lowdelt2,newTau,highbound,lowbound,newdelta,
	startguess,mindelt(*,WExact*),junk1, debug,
	eLocal,dualeigenfunction,dualmaxits,harrissfmaxits,
	harrissftoler,outputchannel,diag,diskmaxits,disktol,
	dim,nsp,perfectsquare,leftdelta,lefttau,bracketok,
	rightdelta,righttau,method,delta0,tau0,nloop,tau,delta	},  



(*. . . PRELIMINARIES: variables given initial values . . .*)

        eLocal = DualCircleGrid         /. {opts} /. Options[DiskIteration];
        dualeigenfunction = DualSingularFunction /. 
	                                   {opts} /. Options[DiskIteration];
        dualmaxits = DualMaxIts         /. {opts} /. Options[DiskIteration];
        dualtol  = DualTol              /. {opts} /. Options[DiskIteration];
	  
	harrissfmaxits=HarrisSFMaxIts   /. {opts} /. Options[DiskIteration];
	harrissftoler= HarrisSFToler    /. {opts} /. Options[DiskIteration];
	outputchannel = DiskIterationOutputChannel/. 
	                                   {opts} /. Options[DiskIteration];
	diag =DiskIterationDiagnostics  /. {opts} /. Options[DiskIteration];
	diskmaxits=DiskIterationMaxIts  /. {opts} /. Options[DiskIteration];
	disktol = DiskIterationTol      /. {opts} /. Options[DiskIteration];
	debug= DiskIterationDebugMe     /. {opts} /. Options[DiskIteration];
        



(*. . . Messages . . .*)
        Debugger::location= " `` ";
        COMMENT[location_]:=
           If[debug, Off[General::stop];Message[Debugger::location,location];
                     On [General::stop]];

        CheckCode::zerodivide= " `` = 0";
        CheckZeroDivide[value_,variable_]:=
           If[value==0., Message[CheckCode::zerodivide,variable];Abort[]];

        CheckCode::withintolerance= " `` ";
        CheckWithinTol[tol_,x_,comment_]:=
           If[x<tol, Message[CheckCode::withintolerance,comment];Abort[]];




(*. . . Initial values . . .*)
	{ndim,nsp} = Dimensions[Ak];
        CheckZeroDivide[nsp,"nsp"];
	If[eLocal==Automatic,eLocal=Table[N[Exp[2 Pi I (i-1)/nsp]],{i,1,nsp}]];
	If[dualeigenfunction==Automatic,dualeigenfunction=Table[1.,{nsp}]];

(* QQ 
	mm = Max[Gk]
	Do[
	If[TrueQ[Sum[Abs[Ak[i,j]],{j,1,ndim}] + mm - Gk[i] < 10^-12], stop,else continue]
	{i,1,ngrid}];
	Do[all entries of Hk should be nonzero ]
*)
	minhk = Min[Chop[Hk]];
	If[minhk==0,
		Print["Hk is 0 at position ",Position[Chop[Hk],minhk]];
		Abort[] ];

(*. . *)COMMENT[" DiskIteration - Complete the square to define W "];
(*. . .           and to determine if the expression is already a perfect square  *);
        squarecomplete = (Gk-Sum[Abs[Ak[[j]]]^2/Hk[[j]],{j,1,ndim}]);

	perfectsquare=
	     If[Chop[squarecomplete.Conjugate[squarecomplete]]==0,True,False];


(*. . . Define K to pass to Nehari . . *)
        K = Table[ Conjugate[Ak[[j]]]/Hk[[j]],{j,1,ndim}];




(*. . *)COMMENT[" Solve quadratic problem:   "]; 

        If[perfectsquare,

(*. . . .*)COMMENT[" perfectsquare True.Bypass inner loop. Go directly to Nehari  "];
          If[diag>=1,WriteString[outputchannel,
		"Subproblem yields a perfect square. Inner loop bypassed"];
          ];


(*. . . .*)COMMENT[" Define W for call to Nehari[[W]]  "];
           WExact = Table[ If[Length[Hk[[i]]]==0, 
                             Flatten[ Table[Hk[[i]],{nsp}] ],
                             Hk[[i]] ],
                    {i,ndim}];

(*. . . .*)COMMENT[" Call Nehari to find update "];
	   {update,junk1,errors,dualeigenfunction,alpha,deiterations}=
               Nehari[WExact,K,
			DualSingularFunction->dualeigenfunction,
			DualMaxIts->dualmaxits,
			DualTol->dualtol,
			DualCircleGrid->eLocal,
			HarrisSFMaxIts->harrissfmaxits,
			HarrisSFToler->harrissftoler,
		        NehariOutputChannel->outputchannel,
		        NehariWrite->TrueQ[diag>=3]];			
	  tau = 1;
	  nloop=0;     
      ,
(*. . . Else . . .*)
(*. . *)COMMENT[" perfectsquare = False "];
(*. . . Since the quadratic expansion of our function is not already a
        perfectwquare, we must attempt to approximate our 
	quadratic expansion. we use an iterative system which intelligently
        picks a 'W' which eventually has the property that W | K + h |^2
        approximates our quadratic expansion 'nicely'.  The approximation is
        considered good when the internal variable "tau" is one.  The variable
        Delta is the variable which is iterated in an attempt to find a proper
        W.  In the above formulation, K is fixed, W depends on delta.*)

          (* . . . Define W for call to Nehari[[W],] . .*)
          CheckZeroDivide[Delta-squarecomplete,"Delta-squarecomplete"];
          W[Delta_] = Table[ Re[Hk[[i]]/(Delta-squarecomplete)], {i,1,ndim}];
          

(*. .. .*)COMMENT[" Bracketing delta "];
          mindelta = Max[Re[Gk - Sum[ (Abs[Ak[[i]]]^2)/Hk[[i]] ,{i,1,ndim}]]];
	  If[diag>=2, WriteString[outputchannel,
	  "mindelta = ",ToString[InputForm[mindelta]]]];
			  

(*. . . .*)COMMENT[" First determine righthandside endpoint (rightdelta,righttau) "];
          rightdelta = Max[Abs[Gk]];
          {update,righttau,errors,dualeigenfunction,alpha,deiterations}=
		Nehari[ W[rightdelta],K,
			DualSingularFunction->dualeigenfunction,
			DualMaxIts->dualmaxits,
			DualTol->dualtol,
			DualCircleGrid->eLocal,
			HarrisSFMaxIts->harrissfmaxits,
			HarrisSFToler->harrissftoler,
		        NehariOutputChannel->outputchannel,
		        NehariWrite->TrueQ[diag>=3]];			

	  If[diag>=2, WriteString[outputchannel,
	  "------------------------\n","Right Tau: ",
	  ToString[InputForm[righttau]],
		             "RightDelta: ",ToString[InputForm[rightdelta]],
			     "\n------------------------"]
          ];


(*. . . .*)COMMENT[" Initial choice of lefthandside endpoint (leftdelta,lefttau)"];
	  leftdelta     = Min[Abs[Gk]];
	  If[leftdelta <= mindelta, leftdelta = 0.5*(mindelta+rightdelta)];
          {update,lefttau,errors,dualeigenfunction,alpha,deiterations}=
		Nehari[ W[leftdelta],K,
			DualSingularFunction->dualeigenfunction,
			DualMaxIts->dualmaxits,
			DualTol->dualtol,
			DualCircleGrid->eLocal,
			HarrisSFMaxIts->harrissfmaxits,
			HarrisSFToler->harrissftoler,
		        NehariOutputChannel->outputchannel,
		        NehariWrite->TrueQ[diag>=3]];
	  If[diag>=2,WriteString[outputchannel,
	  "------------------------\n","Left Tau: ",
	  ToString[InputForm[lefttau]],
	  "LeftDelta: ",ToString[InputForm[leftdelta]],
	  "\n------------------------"]];


(*. . . *)COMMENT[" Loop for bracketing delta that we seek"];
(* . . .            i.e., find leftdelta in [leftdelta,rightdelta] *);
	  nloop = 1;
	  While[ lefttau < 1 && nloop < 10 && (leftdelta-mindelta> 0.00001),
	     rightdelta = leftdelta;
	     righttau = lefttau;
	     leftdelta = mindelta +  2^(-nloop) *(leftdelta - mindelta);
             {update,lefttau,errors,dualeigenfunction,alpha,deiterations}=
		Nehari[ W[leftdelta],K,
			DualSingularFunction->dualeigenfunction,
			DualMaxIts->dualmaxits,
			DualTol->dualtol,
			DualCircleGrid->eLocal,
			HarrisSFMaxIts->harrissfmaxits,
			HarrisSFToler->harrissftoler,
		        NehariOutputChannel->outputchannel,
		        NehariWrite->TrueQ[diag>=3]];			
	     If[diag>=2,WriteString[outputchannel,
	     "------------------------\n","Tau: ",ToString[InputForm[lefttau]],
	                      "     Delta: ",ToString[InputForm[leftdelta]],
			      "\n------------------------"]];
	  nloop++];


	  If[lefttau > 1,
	  	If[diag>=2,WriteString[outputchannel,
		"{leftdelta,rightdelta}={",
		ToString[InputForm[leftdelta]],",",
		ToString[InputForm[rightdelta]],"}"]];
	  	bracketok = True,
		bracketok = False;
		delta = leftdelta;
		tau = lefttau;
	        If[diag>=1,WriteString[outputchannel,
	       "Bracketing delta fails. I will continue with current delta."]];
	  ];
	  

(*. . ..*)COMMENT[" If bracketing was successful, find optimal (delta,tau) "];
	  method = "ParabolaFit";
	  nloop  = 0;
	  delta  = 0.5*(leftdelta+rightdelta);
  	  tau    = 0.5*(lefttau+righttau);      (* to enter the loop *)
          While[bracketok && (nloop<diskmaxits) && (Abs[tau-1]>disktol) && 
	        (delta-leftdelta > 10^-10) && (rightdelta - delta)> 10^-10,


	        method = "ParabolaFit";
		If[method=="ParabolaFit",
(*. . . . . . *)COMMENT["Parabola method begins here "];

                      If[nloop==0,
		         delta = 0.5*(leftdelta+rightdelta)
                      ,
		      (* Else *)
			delta0 = delta;
			tau0 = tau;
                        CheckZeroDivide[delta0-leftdelta,"delta0-leftdelta"];
                        CheckZeroDivide[rightdelta-delta0,"rightdelta-delta0"];

                        If[(tau0-lefttau)/(delta0-leftdelta) < 
			          (righttau-tau0)/(rightdelta-delta0)< 0,
                            (*Then*)
			    c0 = lefttau;
                            CheckZeroDivide[rightdelta-leftdelta,"rightdelta-leftdelta"];
			    c1 = (righttau-lefttau)/(rightdelta-leftdelta);
                            CheckZeroDivide[((delta0-leftdelta)(delta0-rightdelta)
				   (rightdelta-leftdelta)),"c2 devisor"];
			    c2 = ( (tau0-lefttau)(rightdelta-leftdelta)
			          -(righttau-lefttau)(delta0-leftdelta))/
				  ((delta0-leftdelta)(delta0-rightdelta)
				   (rightdelta-leftdelta));
			    parabola = c0 + c1 (x-leftdelta) +
			    		c2 (x-leftdelta)(x-rightdelta);
			    delta=Min[Re[x /. NSolve[ parabola==1.,x]]];
			    If[tau0>1,
				leftdelta = delta0;
			        lefttau = tau0,
			    	rightdelta=delta0;
				righttau=tau0
                            ];

			,
                        (* Else *)
			    method = "Secant";
	                    If[diag>=1,WriteString[outputchannel,
	                   "ParabolaFit fails in finding delta. Switch to Secant method."]];
		        ];

                      ];

		]; (* Parabola method ends here *)



		If[method=="Secant",
(*. . . . . . *)COMMENT["Secant method begins here "];
		      If[nloop==0,
		         delta = 0.5*(leftdelta+rightdelta),
		      (*Else *)
		         If[tau0>1,
        		     leftdelta = delta0;
       		 	     lefttau = tau0,
			 (*Else *)
       		 	     rightdelta = delta0;
       		 	     righttau = tau0;
        	         ];		
                         CheckZeroDivide[
              (righttau-lefttau)+leftdelta,"(righttau-lefttau)+leftdelta"];
		         delta = (1-lefttau)(rightdelta-leftdelta)/
					(righttau-lefttau)+leftdelta
                       ]
                ]; (* Secantmethod ends here *)

                {update,tau,errors,dualeigenfunction,alpha,deiterations}=
		Nehari[ W[delta],K,
			DualSingularFunction->dualeigenfunction,
			DualMaxIts->dualmaxits,
			DualTol->dualtol,
			DualCircleGrid->eLocal,
			HarrisSFMaxIts->harrissfmaxits,
			HarrisSFToler->harrissftoler,
		        NehariOutputChannel->outputchannel,
		        NehariWrite->TrueQ[diag>=3]];			
		
	         If[diag>=2,WriteString[outputchannel,
		 "------------------------\n","Tau: ",ToString[InputForm[tau]],
	                      "     Delta: ",ToString[InputForm[delta]],
			      "\n------------------------"]];

          nloop++; ]; (* end while *)
	        


(*. . ..*)COMMENT[" Check success of final DualExtremal call "];
          (*. . . Did D.E. exceeded iteration number?. . . *)
	  If[deiterations==dualmaxits,
            If[diag>=1, WriteString[outputchannel,
              "*WARN: DualExtremal cannot find subprob. solution in iter. limit"];
            ];
	  ];

          (*. . . Did innerloop reach newtau1?. . .*)
	  If[Chop[newTau - 1,disktol] != 0,
             If[diag>=1,WriteString[outputchannel,"*WARN: InnerLoop did not reach one"]];
	  ];



        ]; (* End of "If perfectsquare" *)




(*. . . Diagnostics . . .*)
(*. . . The error from last call to DualExtremal . . .*)
        If[diag>=1,
          Write[outputchannel,OutputForm["Flatness:          "],errors[[1]]];
          Write[outputchannel,OutputForm["NonAnalyticNoise:  "],errors[[2]]];
          Write[outputchannel,OutputForm["Tail Fou.Coef. K : "],errors[[3]]];
          Write[outputchannel,"  "]
        ];


(*. . *)COMMENT[" End of DiskIteration.  Return final values "]; 
        Return[{update,perfectsquare,errors,dualeigenfunction,
	        nloop,tau}]


   ]; (* end Module DiskIteration *)

(***************************** end DiskIteration ***************************)





(***********************************************************************
*                    Nehari                                      *
************************************************************************
Nehari solves for h in inf_h  sup_e Sum(j) w[j] |k(j) - h(j)|^2.
The weights w[j] are assumed to be positive. 
It solves this by spectral factoring w[j] = Abs[alpha[j]]^2
and rewriting the problem in the form
   inf_hh sup_e  Sum(j) |alpha(j)k(j) - hh(j)| ^2    
By using DualExtremal, one obtains h by h(j) = hh(j)/alpha[j]. 

Examples:
	n = 128;
	e = Table[N[Exp[2 Pi I (i-1)/n]],{i,1,n}];
	k = {1/(e+0.5),1/(e-0.5)};
	w = {Abs[e] + 2, Abs[e+2]};
	test1 := Nehari[w,k,NehariSolverWrite->True, HarrisSFToler->0.000001];
	k1 = 1/(e-0.5);
	k2 = (e^2 + 1)/(e - 0.25);
	w1 = 2. + Re[e];
	w2 = Abs[e^2 + e + 4];
	test2:=Nehari[{w1,w2},{k1,k2}];                                          *
************************************************************************)

NehariOutputChannel::usage=" \n
   Option for Nehari.  Specifies the output channel(s) to be used. ";
NehariWrite::usage=" \n
   Option for Nehari.  True/False, specifies writing or not to an outputchannel";
Nehari::usage=" \n
   Nehari[w,k] solves for h in the supnorm optimization problem \n
         inf_h  sup_e Sum_j w[[j]] |k[[j]] - h[[j]]|^2. \n
   where h is in Hinfinity, e is in the circle, 1 <= j <= ndim, \n
   k is an ndim-long list of discrete complex valued functions, \n
   w is an ndim-long list of discrete positive valued functions.";

Options[Nehari] = 
		{ DualExtremal`DualMaxIts ->15,
		  DualExtremal`DualTol -> 0.0000001,
		  DualExtremal`DualCircleGrid ->Automatic,
		  DualExtremal`DualSingularFunction->Automatic,
		  AnoptUtilities`HarrisSFMaxIts->10,
		  AnoptUtilities`HarrisSFToler->1.0,
		  NehariOutputChannel->OutputStream["stdout",1],
		  NehariWrite->True }


			    
Nehari[w_,k_,opts___Rule]:=
   Module[{junk1,junk2,AlphaK,er1,er2,er3,er4,er5,er6,j,
	  function,tau,dualeigenfunction,alpha,deiterations,
	  eLocal,dualmaxits,dualtol,harrismaxits,specfactol,
	  outputchannel,write,ndim,nsp},


        eLocal = DualCircleGrid       /. {opts}/.Options[Nehari];
        dualeigenfunction = DualSingularFunction /. 
	                                 {opts}/.Options[Nehari];
        dualmaxits = DualMaxIts       /. {opts}/.Options[Nehari];
        dualtol  = DualTol            /. {opts}/.Options[Nehari];
	  
	harrismaxits = HarrisSFMaxIts /. {opts}/. Options[Nehari];
	specfactol   = HarrisSFToler  /. {opts}/. Options[Nehari];
	outputchannel = NehariOutputChannel/. 
	                                 {opts}/. Options[Nehari];
	write = NehariWrite           /. {opts}/. Options[Nehari];


	{ndim,nsp} = Dimensions[k];
	If[eLocal==Automatic,
		eLocal=Table[N[Exp[2 Pi I (i-1)/nsp]],{i,1,nsp}]];
	If[dualeigenfunction==Automatic,
		dualeigenfunction=Table[1.,{nsp}]];

 

(*. . *)COMMENT[" Nehari begins: Spectral factorization "];
(*. . . Construct alpha subject to (alpha)^2 = w.  Two routines may be
        called: AnoptUtilities`Wilson and AnoptUtilities`Harris. . . *)


(*. . *)COMMENT[" Call WilsonSF "];
        {alpha,er5,er6} = 
	   Transpose[ {SpectralFactor,SpectralFactorError,
                       SpectralFactorNonAnalyticNoise }/. 
                    Table[
                       WilsonSF[w[[j]],
                    	   WilsonSFSave->write,
                    	   WilsonSFOutputChannel->outputchannel]
              	    ,{j,1,ndim}]
	   ]; 
        er5 = Apply[Plus,Abs[er5]];  
	er6 = Apply[Plus,Abs[er6]];  



(*. . *)COMMENT["  Call HarrisSF "];
(*. . . If er5+er6 from Wilson is too large, re-spectral factorize.
        To avoid HarrisSF, just set specfactol to be a large number  . .*)

	If[(er5+er6>specfactol),
             {alpha,er5,er6,junk1,junk2} =
                Transpose[{SpectralFactor,SpectralFactorError, 
                          SpectralFactorNonAnalyticNoise,
                          HarrisSFStep, HarrisSFIterations}/.
                     Table[ HarrisSF[w[[j]],
			    HarrisSFToler->specfactol,
                            HarrisSFSave->write,
                            HarrisSFOutputChannel-> outputchannel]
                     ,{j,1,ndim}]
             ];  
             er5 = Apply[Plus,Abs[er5]];  
	     er6 = Apply[Plus,Abs[er6]];  
        ];



(*. . *)COMMENT[" Call DualExtremal:  Solve for H "];

(*. . . AlphaK is alpha found by multiplying alpha*k term by term . ..*)
	AlphaK = Partition[Flatten[alpha] Flatten[k], nsp]; 

        (****** CALL DualExtremal ******)
        {tau,{er2,er1,er4,er3,deiterations},function,dualeigenfunction}=
            {Tau,DualOutputDiagnostics,-HInfinityApprox,DualSingularFunction}/.
	       DualExtremal[ AlphaK, 
                     DualCircleGrid ->eLocal,
                     DualSingularFunction->dualeigenfunction,
                     DualMaxIts->dualmaxits,
                     DualTol->dualtol,
                     DualExtremalSave->write,
                     DualOutputChannel->outputchannel,
                     DualDebugMe->debug ];
	function=Table[function[[i]]/alpha[[i]],{i,1,ndim}];




(*. . *)COMMENT[" End of Nehari:  Return final value "];
	Return[{function,tau,{er1,er2,er3,er4,er5,er6},
	        dualeigenfunction,alpha,deiterations}]

   ] (* end  Nehari Module*)
(******************** end Nehari **********************************)


End[];  (* "`Private`" *)

EndPackage[];
