(* :Title: Dual Extremal*)

(* :Author: Orlando Merino, programming.
	    Algorithm by D. Schwartz and J.W. Helton. *)

(* :Summary: 
    Finds the highest singular value and associated vector of the 
    hankel operator H_K with symbol K = a function on the unit circle,
    using the power method applied to  the matrix Conjugate[Transpose[H_K] . H_K
    Here H_K can be a single hankel matrix -if the function K is scalar-
    or a vector with entries which are Hankel matrices -if the function
    K is vector valued.
*)

(* :Context: DualExtremal` *)

(* :Package Version: 1.2 *)

(* :History:
    Standardized internal documentation by Julia Myers, September 1990. 
    Version 1.1 by Orlando Merino, July 1990.
    Version 1.2 by O.M. Dec.92
*)

(* :Keywords: BestHinfinity approximation, Dual Extremal method *)

(* :Source:
  - Helton, J. W., Schwartz, D. "Best Uniform Approximation of 
    Vector Valued Functions by Hoo functions."  Preprint.
  - Adamjan, Arov, and Krein. ??
  - Hintzmann, "Best Uniform Approximation of Functions via
    Anihilating Measures."  Bull. AMS ??
*)

(* :Warning: *)

(* :Mathematica Version: 2.0 or higher *)

(* :Requirement: *) 

(* :Limitation:
    The following conditions must be satisfied for the algorithm to work:
    - n must be > 8
    - The tail of Fourier coefficients of K is small when compared to gridsize.
    - The multiplicity of the highest singular value of
      k1* k1 + ... + k_dim* k_dim  is equal to 1 (this holds for generic K)
      and there are no other singular values "near" the largest one.
*)

(* :Discussion: 
    The Function DualExtremal solves the discrete version of the problem:
    GIVEN a function K from the unit circle to C^N
    FIND  tau, the distance from K to Hoo, and the best approximation
   or     to K from Hoo.
    FIND  tau, the norm of the Hankel operator H_K of symbol K, and a scalar
          function x in H2 where the norm of H_K is attained, in particular
	  the following equation must be satisfied:
	         Conjugate [P[K e P[Conjugate[K e x ]]]  =  tau^2 Conjugate[e x]
          Here P is the orthogonal projection onto the perp of H2.

    Algorithm: 
    Starting with a given K = { k1 , k2 , ... , k_dim },
    and a function x 
    - Compute y = Conjugate [P[K e P[Conjugate[K e x ]]]
    - Obtain a current value for tau from  tau^2 = SupNorm[y]/SupNorm[x]  
    - Update x through the equality  x = (1/tau^2 ) y
    - Stop when change in x is smaller than a given tolerance.
        
	Speed of convergence depends on the separation on the two largest
	singular values.  Accuracy depends on the size of grid and tolerances.
*)

(*  :Example: 

	e = Table[N[Exp[2 Pi I (i-1)/128]],{i,1,128}];
	k = { 0.2 /e + 1.5 /e^2 -3.1/e^3 };
	DualExtremal[k] (* here DualExtremal[k,DualCircleGrid->e] also works *)

*)

BeginPackage[ "DualExtremal`" ];

DualExtremal::usage = " \n
   DualExtremal[K,opts ] finds the largest singular value of the operator \n
   with symbol K, where K = {K1 , K2 , . . .} and Kj is a discrete function \n
   (i.e., a list of samples) on a grid of equally spaced points on the \n
   unit circle in the complex plane."
Tau::usage = " \n
   Tau is output of DualExtremal, the Norm of Hankel operator with symbol K."
DualSingularFunction::usage = " \n
   DualSingularFunction is output to DualExtremal.  It is the function at  \n
   which the Hankel operator with symbol K attains its norm. It can be an \n
   input also, used by DualExtremal to initialize the iteration."
HInfinityApprox::usage = "\n
   HInfinityApprox is a discrete function, the best Hinfinity approximation \n
   to K.  It is an output of DualExtremal."
DualOutputDiagnostics::usage = " \n
   DualOutputDiagnostics is output of DualExtremal, a list of real numbers \n
   communicated as DualOutputDiagnostics -> { od1, od2 , od3 }.  Here, \n
   od1 = Max [Abs[K - f]] - Min[Abs[K-f]], where f is the HInfinityApprox. \n
   od2 = Sum[Abs[fourier coeffs. of f with neg. index]], \n
         where f=HInfinityApprox \n
   od3 = Sum[Abs[fourier coeffs. of K with index > 7 n/8 or < -7 n/8 ]]. \n
   od4 = QQ
   od5 = QQ
   Small odj,j=1,3 corresponds to a successful run." 
DualExtremalSave::usage = " \n
   DualExtremalSav is an Option for DualExtremal[].  When set to True \n
   a run of DualExtremal prints diagnostics that show progress of the \n
   iteration.  The channel for printing is specified with DualOutputChannel."
DualTol::usage = " \n
   DualTolOption is an input of DualExtremal[].  The tolerance used to stop \n
   the loop iteration."
DualMaxIts::usage = "\n
    DualMaxIts is an option for DualExtremal[].  It limits the number of \n
    iterations of the main loop."
DualCircleGrid::usage = " \n
   DualCircleGrid is an Option for DualExtremal[].  Specifies a grid of \n
   equally spaced points on the unit circle in the complex plane.  Used \n
   only when this grid is already available (this is to save computer time),\n
   and it must agree with the target function K on number of samples."
DualOutputChannel::usage = " \n
   DualOutputChannel is an option for DualExtremal[].  Specifies the output\n
   stream for printing error diagnostics."
DualDebugMe::usage=" \n"
Debugger::location::usage = " \n
    Message for DiskIterationDebugMe option.  Function COMMENT[] controls it "
   

Options[DualExtremal] = {DualExtremalSave -> True, 
                         DualTol         -> 10^(-10),
                         DualMaxIts      -> 50,
			 DualOutputChannel-> OutputStream["stdout",1],
			 DualSingularFunction  -> Automatic,
			 DualCircleGrid -> Automatic,
                         DualDebugMe->False};
                      
Begin["`Private`"];

DualExtremal[ K_(*List*),opts___Rule] :=
   Module[ {ik, i, j, delta, zero, n, dim, zpr, y, y1, y2, tau2, kNew,
           kfcNew, e, xOne, iv, deltaTol, maxits,channel, debug },

   (*    initialize variables   ................................  *)
   
      iv       = DualExtremalSave     /. {opts} /. Options[DualExtremal];
      channel  = DualOutputChannel    /. {opts} /. Options[DualExtremal];
      deltaTol = DualTol              /. {opts} /. Options[DualExtremal];
      maxits   = DualMaxIts           /. {opts} /. Options[DualExtremal];
      e        = DualCircleGrid       /. {opts} /. Options[DualExtremal];
      xOne     = DualSingularFunction /. {opts} /. Options[DualExtremal];
      debug    = DualDebugMe          /. {opts} /. Options[DiskIteration];
      If[TrueQ[e==Automatic],e = N[ Table[Exp[2 Pi I(i-1)/n],{i,1,n}]]];
      If[xOne == Automatic, xOne = Table[1.,{j,1,n}] ];

      {dim,n}  = Dimensions[K];
      zero    = Table[0.,{i,1,n/2+1}];
      delta = 1.;


(*. . . Messages . . .*)
        Debugger::location= " `` ";
        COMMENT[location_]:=
           If[debug, Off[General::stop];Message[Debugger::location,location];
                     On [General::stop]];


(*. . *)COMMENT[" Dual Extremal Begins:"];

(*. .  Zero-out Fourier Coefficients of non-negative index in K 
      to get the projection of K onto H^{2 perp}.  We work with this K *) 
      
      kNew   = Table[ Fourier[K[[i]]],{i,1,dim}];
      kfcNew = Table[ Join[{0.}, Take[kNew[[j]],{2,n/2}],
               Table[ 0.,{i,1,n/2}]],{j,1,dim}];
      kNew   = Table[ InverseFourier[kfcNew[[j]]],{j,1,dim}];


   (*    Heading for optional diagnostics  ................ *)
      If[ iv, Write[channel,OutputForm[" "]];
              Write[channel,OutputForm["DualExtremal DIAGNOSTICS OUTPUT "]];
      	      Write[channel,
                OutputForm[ "Iter.  Change in SingularVector        Tau2"]]
      ];


(*. . *)COMMENT[" Main Loop"];
(* ...............................................  *)
      ik = 1;
      While[ (delta > deltaTol)  &&  (ik <= maxits),

         Do[    
            zpr   = Join[ Take[ Fourier[kNew[[i]] xOne], {2,n/2} ],zero];

            y1[i] = Conjugate[ InverseFourier[zpr]];
            zpr   = Join[ Take[Fourier[kNew[[i]] y1[i] ],{2,n/2}],zero];
            y2[i] = Conjugate[ InverseFourier[zpr]];
         , {i,1,dim} ]; (* end Do *)

         y     = Sum[ y2[i], {i,1,dim}] ;
         tau2  = Max[ Abs[y] ];
         delta = Max[ Abs[xOne - y/tau2 ] ];
         xOne  = y/tau2;
         If[iv, Write[channel,ik,OutputForm["      "],
                     delta,OutputForm["            "],N[tau2,16]] ];
         ik++;
      ]; (* end While *)

(*. . *)COMMENT[" End of Main Loop"];

   (*  Final output *)
      tau             = N[Sqrt[tau2],16];  
      hinf  = Table[ K[[i]] - Conjugate[e y1[i] ]/ xOne,{i,1,dim}];

(*. . *)COMMENT[" Error diagnostics calculation - output diagnostics"];
      od1 = Max[ Sum[Abs[K[[i]]-hinf[[i]]]^2,{i,1,dim}]] -
             Min[ Sum[Abs[K[[i]]-hinf[[i]]]^2,{i,1,dim}]]; 
      od2 =Sum[Sum[Abs[Fourier[hinf[[i]]][[j]]],{j,2,n/2}],{i,1,dim}]
      			/Sqrt[N[n]];
      od3 = Sum[Apply[Plus,Abs[ Take[Fourier[K[[i]]], {3 n/8,5 n/8}] ]]
      			,{i,1,dim}]/N[Sqrt[n]] ;
      od4 = delta;
      od5 = ik - 1;
      
      DualOutputDiagnostics={od1,od2,od3,od4,od5};
      If[iv,Write[channel,OutputForm["{DualOutputDiagnostics}->"],
                  DualOutputDiagnostics]];

(*. . *)COMMENT[" End of DualExtremal:  Return final values"];
      Return[ {Tau -> tau,
            DualOutputDiagnostics -> {od1,od2,od3,od4,od5},
            HInfinityApprox -> hinf, 
	    DualSingularFunction -> xOne}];

   ]; (* end DualExtremal Module *)
         

End[(* "`Private`" *)];

EndPackage[];
