(* --------------------------------------------------  
		 Package NewtonFit 
 
Laboratory for Mathematics and Statistics, 
University of California, San Diego. 
 
Copyright J. Easton and O. Merino (c) 1991,2001
All rights reserved.  The usual disclaimer applies. 
 
---------------------------------------------------- *) 
(* :Title: NewtonFit *) 
 
(* :Authors:  
	Programming: Jim Easton and Orlando Merino. 
	Algorithm:   Newton's method for L2 approximation 
*) 
	 
 
(* :Summary: ??  *)   
(* 	This is an implementation of the  -Newton algorithm 
	adapted to find local best rational fit of given order 
	to given data in the L2 norm. 
*) 
	 
(* :Context: NewtonFit` *) 
 
(* :Package Version: 0.1 May 1991. *) 
 
(* :History: 
    Experimental version 0.1 by J.Easton  and O.Merino, May 1991. 
		Rational Function Approximation only.
    version 0.2 by O. M. , June 1991.
		General nonlinear approximation.
    version 0.3 by O. M. , January 1992.
	        Golden section line search added.
		Output format was modified.
		Minor changes in documentation of Options.
*) 
 
(* :Keywords: NonLinear Modelling, Least Squares Fit *) 
 
(* :Source: ?? *) 
 
(* :Warning: *) 
(* 
	This program does not make any assumptions 
	about the symmetry of the function being approximated. 
	If one exists, the input data must reflect it. 
 
*) 
 
 
(* :Mathematica Version: 1.2 *) 
 
(* :Requirement: none *)  
 
(* :Limitation: ?? *) 
 
 
 
 
 
(* :Discussion:   
 
    THE PROBLEM. 
 
	GIVEN 
		npoints, an integer 
		points, a list of complex numbers of length=npoints 
		values, a list of complex numbers of length=npoints 
		weight, a list of positive numbers of length=npoints
		model, a function of parameters a[1],...,a[m] 
		 
	FIND COMPLEX NUMBERS  {a1,...,am} 
 
	THAT MINIMIZE (LOCALLY) THE OBJECTIVE FUNCTION 
 
	Chi2 = Sum[weight[[i]]*Abs[values[[i]]-model[points[[i]] ]]^2 
					,{i,1,npoints}] 
	 
	WHERE 
 
	model = a function of x and parameters a[1],a[2],...
 
    Algorithm:  To update a[] to a'[] do: 
	 
	1)Solve the Newton Step equation Chi2[a] + Chi2'[a,deltaa] == 0.
	  This gives the update direction deltaa.
 
	2)Pick a t_0 >0 such that a[] + t_0 *deltaa[] minimizes 
	  chi2 over all t>0. 
 
	3)Set a'[] = a[] + t_0 * deltaa[] 
	 
    Implementation: 
 
    Step 1. Nothing to say here.  Pretty straightforward. 
    Step 2. golden section line search is performed. 
 
    Stopping.  The iteration stops if either  
 
	a. the chi2 function is close to 0 within a given tolerance. 
	b. the number of iterations reaches a prescribed maximum. 
	c. the gradient of the function is close to 0 
		within the given tolerance. 
	d. the step size=euclidean norm of difference of iterates is 
		close to 0 whithin tolerance. 
*) 
 
 
 
 
 
 
 
 
 
 
 
 
BeginPackage["NewtonFit`"] 
 
(* usage messages for the exported functions and the context itself *) 
 
	Maxits ::usage="Option for NewtonFit, a positive integer. 
	It specifies the maximum number of iterations allowed. ";
 
	Tolerance::usage="Option for NewtonFit.  A positive number, 
	the iteration in NewtonFit is stopped if the change in the 
	parameters is smaller than Tolerance. ";
 
	Diagnostics::usage="Option for NewtonFit, a logical variable. 
	If True, then indicators of progress of the iteration are  
	printed on the screen. "; 
 
	Weight::usage="Options for NewtonFit, a list of positive numbers. 
	It is part of the formula of the objective function.  It could carry
	information on standard deviation of error at each data point,
	or it could be used to emphasize certain data points. 
	Default is {1,1,...,1} "; 
 
	Cost::usage="Output to NewtonFit, a real number.  The value of
	the objective function at the computed parameter values. ";
 
	Parameters::usage=" List of parameter values.  It can be specified as
	an option to initilize computation in NewtonFit.  If it is not 
	given as input, then it is defauted to {0.,0.,...}.  It is also an
	output, the name of the computed list of parameters. ";
 
	StepSize::usage="Output to NewtonFit, the Euclidean norm of the 
	difference of the last two iterates in the parameters. "; 
 
	Gradient::usage="Output to NewtonFit, the gradient in a[1],a[2],.   
	of the objective function at the computed solution "; 

	NormGradient::usage="Output to NewtonFit, the norm of the
	gradient in a[1],a[2], of the objective function at the computed
	solution ";
	
	SingValHessian::usage="Singular values of the hessian of the \n
	cost function at the computed solution.";
 
	NIterations::usage="Output to NewtonFit, the total number of  
	iterations required in the computation."; 
  
	NewtonFit::usage="   \n
	NewtonFit[data,model,var,param]  produces \n
	(1) a list of parameter values that minimize in a[1],a[2],... \n
	   (locally) the chi square function given by \n
	 \n 
	    Sum[  wi* Abs[yi-model[xi,a[1],...a[n]] ]^2 , {i,1,nx} ] \n 
	\n
	   where the wi's are defaulted to 1 (see option Weight). \n
	(2) additional output wich gives information on the run \n
	Input: \n 
		data    a list {{x1,y1},{x2,y2}, . . . } \n
		model   a differentiable function of var and param[1],param[2],. . . \n
		var	the name of the independent variable \n
		param	the name of the parameters.  ";
 
 
 
Options[NewtonFit]={Maxits->15,Tolerance->10^(-10), 
			Parameters->Automatic,Weight->None,
			Diagnostics->False} 
 
 
Begin["Private`"]; 
 
 
NewtonFit[data_,model_,var_,param_,opts___Rule]:= 
Block[{i,j ,npoints,ordern,orderd,tolerance,diag, maxits,weight0, 
	zn,zm,matrix,niter,errorcurve, 
	chi2,jmatrix,jf,rational,denrational,gradmat, 
	gradient,normgradient,part1,part2,chi2table, 
	minchi2,imin,tmin,x,a}, 
(* check for incorrect input *) 
 
 
(* initialize local variables *)
	localmodel = model /. {var->x,param->a}; 
	{datapoints,datavalues}=Transpose[data];
	npoints = Length[datapoints];
	one = Table[1.,{npoints}]; 
	order = 1; 
	While[TrueQ[D[localmodel,a[order+1]]=!=0],order++]; 
	grad = Table[D[localmodel,a[i]],{i,1,order}];
	hess = Table[D[grad[[i]],a[j]],{i,1,order},{j,1,order}]; 
	diag = Diagnostics /. {opts} /. Options[NewtonFit]; 
	maxits = Maxits /. {opts} /. Options[NewtonFit]; 
	tolerance = Tolerance /. {opts} /. Options[NewtonFit]; 
	tlist = Join[Table[0.1 i,{i,0,15}],Table[2.^i,{i,1,4}]]; 
	stepsize = 10^10; 
	niter =0; 
	weight0 = Weight /. {opts} /.  Options[NewtonFit];
	If[TrueQ[weight0 == None], weight0 = one]; 
	ak = Parameters /. {opts} /. Options[NewtonFit];
	If[TrueQ[ak == Automatic],ak=Table[0.,{order}]];
	If[Length[ak]==0,ak = Table[0.,{i,1,order}]];
	tmodelk[t_] :=  localmodel /.{a[j_]:>ak[[j]]+t*adelta[[j]], 
					x->datapoints}; 
	Fline[t_]:= (Abs[(datavalues - tmodelk[t])]^2 ).weight0;
(* check for incorrect input in Options *)  
	If[Length[ak]==order, Null,
		Print["Number of parameters in model is not the same \n
			as the number of parameters given as input "];  
		Return[] ];	
	If[Length[weight0]==npoints, Null, 
		Print["Length[Weight] not equal to Length[f]"]; 
		Return[] ]; 
(* Begin computations.  Compute initial least-squares error *) 
	modelk = localmodel /.{a[j_]:>ak[[j]],x->datapoints};
	If[diag,Print["dim datavalues = ",Dimensions[datavalues]]];
	If[diag,Print["dim modelk = ",Dimensions[modelk]]];
	chi2 =  (Abs[(datavalues - modelk)]^2).weight0;
	gradk = Table[
		  If[TrueQ[D[grad[[i]],x]==0],  
			Table[grad[[i]],{npoints}],  
			grad[[i]] /.{a[j_]:> ak[[j]],x->datapoints}], 
				{i,1,order}];
	hessk = Table[  
		  If[TrueQ[D[hess[[i,k]],x]==0],  
			Table[hess[[i,k]],{npoints}],  
			hess[[i,k]] /.{a[j_]:> ak[[j]],x->datapoints}], 
				{i,1,order},{k,1,order}];
	costgrad = (Table[
		2. Re[ Conjugate[datavalues-modelk]*(-gradk[[j]])],
		   {j,1,order}]).weight0;
	costhess = (Table[ 2. Re[Conjugate[gradk[[j]]] gradk[[k]]+
			Conjugate[datavalues-modelk] (-hessk[[j,k]])],
		    {j,1,order},{k,1,order}]).weight0;
	normcostgrad = Sqrt[Sum[ Abs[costgrad[[i]]]^2,{i,1,order}]];
	If[diag,Print["costgrad = ",costgrad]];
	If[diag,Print["costhess = ",costhess]];
	If[diag,Print["cost     = ",chi2]];
	unkb = Table[b[i],{i,1,order}];
 
(* print header for screen output *) 
		Print["--------------------------------------------------"]; 
		Print["Iter    step          grad           cost"];
		Print["--------------------------------------------------"]; 
		Print[niter,"                   ",
			normcostgrad,"         ",chi2];
 

(* begin main loop *) 
	While[(niter<maxits)&&(stepsize > tolerance)&& 
			(chi2 >tolerance)&&(normcostgrad > tolerance), 
		niter++; 
		ak1 = ak;
		sv = SingularValues[costhess][[2]];
(* If a singular value is 0, mathematica chops the list of singular values;
   thus we have to restore the list to the correct length *)
		If[Length[sv]<Length[costhess],
		   sv = Join[sv,Table[0.,{Length[costhess]-Length[sv]}]];];
		smallsv = Min[sv];
		largesv = Max[sv];
		If[diag,Print["costhess = ",costhess]];
		If[diag,Print["costgrad = ",costgrad]];
If[diag,Print["smallsv = ",smallsv]];
If[diag,Print["largesv = ",largesv]];
If[diag,Print["sv = ",sv]];
(* If the smallest singular value is too small, 
   take some action to prevent numerical problems *)
		lambda = If[smallsv/largesv < 10^(-4),
				(10. Abs[smallsv]+0.01 ),0.];
If[diag,Print["lambda = ",lambda]];
(* HERE IS NEWTON'S STEP EQUATION *)
		adelta = unkb /. 
		    Solve[ 
		     (costhess+lambda*IdentityMatrix[order]).unkb 
		     	== -costgrad,unkb][[1]];
		
(* LINE SEARCH:  compute scalar t=tmin that minimizes chi2 *) 
	If[diag,Print["adelta = ",adelta]];
		{tk,chi2}=LineSearch[Fline];



	(* update parameter values and compute useful indicators *) 
		ak = ak1 + tk*adelta; 
		gradk = Table[  
		  If[TrueQ[D[grad[[i]],x]==0],  
			Table[grad[[i]],{npoints}],  
			grad[[i]] /.{a[j_]:> ak[[j]],x->datapoints}], 
				{i,1,order}]; 
		modelk = localmodel /.{a[j_]:>ak[[j]],x->datapoints};
		
		hessk = Table[  
		  If[TrueQ[D[hess[[i,k]],x]==0],  
			Table[hess[[i,k]],{npoints}],  
			hess[[i,k]] /.{a[j_]:> ak[[j]],x->datapoints}], 
				{i,1,order},{k,1,order}];
		costgrad = (Table[
			2. Re[ Conjugate[datavalues-modelk]*(-gradk[[j]])],
		   	{j,1,order}]).weight0;
		costhess = (Table[ 2. Re[Conjugate[gradk[[j]]] gradk[[k]]+
				Conjugate[datavalues-modelk] (-hessk[[j,k]])],
		    	{j,1,order},{k,1,order}]).weight0;
		normcostgrad = Sqrt[Sum[ Abs[costgrad[[i]]]^2,{i,1,order}]];

If[diag,Print["adelta = ",Sqrt[Apply[Plus,Abs[adelta]^2]]]]; 
		stepsize = tk*Sqrt[Apply[Plus,Abs[adelta]^2]]; 
(* if diagnostics requested, compute and print at each iteration *) 
		Print[niter,"   ",stepsize,"         ",
				normcostgrad, "         ",chi2];
	    ]; (* end While *) 
(* HERE *)
(* compute final error diagnostics, if not already done *) 
 	    If[diag,Print["At end of run:  "];
	    		Print[" Gradient->",costgrad]];	
(*		output = model /. {param[j_]:>ak[[j]]};	*)
		output = {Parameters->ak,
			  SingValHessian->sv,
			  Cost -> chi2,
			  NormGradient -> normcostgrad,
			  NIterations->niter
			  }; 
		Return[output]; 
]; (* end of Block *) 
 
 
 
 




(* January 1, 1992.  O. Merino. *)
(* This is a linesearch on the semixes 0<x,
   a real valued function F[t] of the real variable t is given 
   by the user. 
   Some of this code was taken from Numerical Recipes *)
LineSearch[F_]:=
    Block[{xa,xb,xmin,fmin}, 
	{{xa,fa},{xb,fb}} = BracketMin[F];
	If[Abs[xa-xb]<0.0002,Return[{xa,fa}]];
	{xmin,fmin} = GoldenSection[F,xa,xb];
	If[Abs[xa-xmin]<0.001,Return[{xa,fa}]];
	Return[{xmin,fmin}];
     ];

BracketMin[F_]:=
    Block[{val0,val,larget,factor,t,cond,difft,oldval,oldt},
	val0 = F[0.];
	val1  = F[1.];
	val  = Min[val0,val1];
	larget = TrueQ[val<val0];
	factor = If[larget,2.,0.5];
	t = 1.;
	cond = True;
	difft = 1.;
	While[cond&&(difft>0.0001)&&(difft<10000.),
	(*	Print[" t = ",t," val = ",val]; *)
		oldval = val;
		oldt = t;
		t = t*factor;
		val = F[t];
		cond = TrueQ[val<oldval];
		If[(!larget)&&val>val0,cond=True];
		difft= Abs[t-oldt];
	     ];
	If[larget, Null,
		(* else *)
		t = oldt;
		val = oldval;];
	Return[{{0.,val0},{t,val}}];
     ];

GoldenSection[F_,xLeft_,xRight_]:=
   Block[{tol=0.0001,gold,gold2,x0,x1,x2,x3,f1,f2},
	gold = 0.5*(-1.+Sqrt[5.]);
	gold2 = 1. - gold;
	x0 = xLeft;
	x3 = xRight;
	x1 = x0+gold2*(x3-x0);
	x2 = x0+ gold*(x3-x0);
	f1 = F[x1];
	f2 = F[x2];
	While[Abs[x3-x0]> tol*(Abs[x1]+Abs[x2]),
	(*
		  Print["{x0,x1,x2,x3} = ",x0," ",x1," ",x2," ",x3];
		  Print["{f1,f2} = ",f1," ",f2];
	*)
  
		If[f2<f1,
			x0=x1;x1=x2;x2=gold*x1+gold2*x3;
			f1=f2;f2=F[x2],
		(* else *)
			x3=x2;x2=x1;x1=gold*x2+gold2*x0;
			f2=f1;f1=F[x1];
		   ];
	      ];
	If[f1<f2,Return[{x1,f1}],Return[{x2,f2}]];
       ];

End[];  (* end the Private context *) 

EndPackage[]; (* end the Package context *) 





test := ( 
random = {0.5874338872666333, 0.7775251076821638, 0.3467599757149066, 
   0.5148344839606417, 0.18065305921487, 0.92289575944166, 
   0.06027727549504812, 0.935495950683088, 0.6775684045183425, 
   0.6803018771302835};
data = Table[{i,2*i+3 + random[[i]]},{i,1,10}]; 
model = (a[1]*x + a[2])/(10.+a[3]*x); 
NewtonFit[data,model,x,a]); 
