(*	:Title:	NewtonInterpolant 
	:Authors: Orlando Merino, Daniel Lam.  Version 1 (1990)
		  O. Merino. Version 2 (Feb. 1993)
	:References: 
		Classical Control with Hinfinity Methods. 
		By J. W. Helton and O. Merino.
		
		Numerical Methods, by C. deBoor and R. Conte.
		
	:Summary:
		Constructs an interpolating rational fn. by
		using the Newton algorithm.
		
		The interpolating rational function  is
		constructed as a sum of terms each of the form 
			coefficient*product*(1/(zz - a )^r) 
		each product has the form 
			(zz-point)/(zz-a) *...* (zz-point)/(zz-a)       

	:Package Version: 2.0 
	
	:Examples:
		
data = { {3,1} , {1,-2} };
rat1 = NewtonInterpolant[data,s]

rat2 = NewtonInterpolant[data,s,PoleLocation->-4]

rat2 = NewtonInterpolant[data,s,PoleLocation->-4,RelativeDegree->2]
								
data = { {I,0},{-I,0},{2 + 3 I,4 - 2 I},{2 - 3 I,4 + 2 I},{1,3} };
rat4 = NewtonInterpolant[data,s]

data = { {1,-3} , {2,0,1} };
rat = NewtonInterpolant[ data, s]

data = { {2 I,0,1 - 3 I},{-2 I,0,1 + 3 I} };
NewtonInterpolant[ data ,s, RelativeDegree -> 1] 

*)

BeginPackage["NewtonInterpolant`"];

   NewtonInterpolant::usage="NewtonInterpolant[data,var],NewtonInterpolant[data,var,opts], produces a rational function in the variable `var', that satisfies interpolation conditions specified in the input `data';  here data = {data1,data2,...} where data_k = {x0,y0,y1,...} and x0 = base point, y0 = value of function at x0, y1 = value of derivative[1] at x0, etc.";

RelativeDegree::usage="Option for NewtonInterpolant.  Specifies the relative degree of the interpolant.";

PoleLocation::usage="Option for NewtonInterpolant.  Specifies the location of the only pole of the interpolant.";

Options[NewtonInterpolant]={PoleLocation->-1,RelativeDegree->0};

Begin["`Private`"];
	
   NewtonInterpolant[data_List,var_,opts___Rule]:=
	Module[{product,currentinterpolant,i,j,keyeqn,coef,zz,r,aa},
	     r =  RelativeDegree /. {opts} /. Options[NewtonInterpolant];
	     aa = PoleLocation   /. {opts} /. Options[NewtonInterpolant];
	     currentinterpolant=0;
	     product=1/(zz-aa)^r;    (* recall r= relative degree *)
	     i = 0;
	     While[i<Length[data],
	     	i++;
		j = 0;
(* The next loop for fixed i enforces the constraints
   D[f,{zz,j-1}] at data[[i,1]] = data[[i,j+1]]  *)
	     	While[j<Length[data[[i]]]-1,
		     j++;
		     keyeqn=D[currentinterpolant+coef*product,
				{zz,j-1}] == data[[i,j+1]];
		     ans=Solve[(keyeqn /. zz->data[[i,1]]),coef];
		     currentinterpolant=
		         currentinterpolant+coef*product /. ans[[1]];
(* this is the next summand but without the coefficient	*)
		     product	=product*(zz-data[[i,1]])/(zz-aa)
		     ]  (* end While *)
		    ];  (* end While *)
	
(*  If currentinterpolant = 0, reset it as  product.  *)
	If[Chop[currentinterpolant] == 0, currentinterpolant = product];
	
	     Return[ Together[currentinterpolant] /. zz->var];
	      ] (* end of Block *) 
 
 End[] (* `Private` *)

 EndPackage[] 

